#import grpc-package
from grpc import insecure_channel

#import class Sequence and OutoutState for advanced sequence building
from pulsestreamer.sequence import Sequence, OutputState

try:
    #import the python wrapper generated by the protocol buffer compiler (source: pulse_streamer.proto)
    from pulsestreamer.grpc.pulse_streamer_pb2_grpc import PulseStreamerStub
    from pulsestreamer.grpc.pulse_streamer_pb2 import VoidMessage, PulseMessage, SequenceMessage, ClockMessage, DrpClkSetMessage, TriggerMessage, SetNetworkMessage, GetNetworkMessage, HostnameMessage, SetSquareWaveMessage, AnalogCalibrationData, SupplyStatusData, StartMessage, ReadyForDataMessage
    from pulsestreamer.grpc.pulse_streamer_pb2 import GetSerialMessage # used for backwards compatibility reasons if client software detects firmware version < 1.1.0
except Exception as e:
    print('Exception: '+ str(e))
    assert False, \
"""
Failed to import grpc wrapper generated by the protobuf compiler
Required packages: grpcio, protobuf
Ensure that you have it installed by typing
> pip install grpcio or pip install --upgrade grpcio
> pip install protobuf or pip install --upgrade protobuf
"""

from pulsestreamer import __version__
from pulsestreamer.enums import ClockSource, TriggerStart, TriggerRearm, NextAction, When, OnNoData
from pulsestreamer.version import  _compare_version_number

#binary conversion
import numpy as np

#stream binary
import socket
import struct

class PulseStreamer():
    """
    Simple python wrapper for a PulseStreamer 8/2
    that describes sequences in the form of sequence steps as (time, [0,1,3], 0.8, -0.4),
    where time is an integer in ns (clock ticks),
    [0,1,3] is a list numbering the channels that should be high
    the last two numbers specify the analog outputs in volt.
    For advanced sequence creation use the method createSequence() and the functionality of
    the class Sequence described in the documentation of the Pulse Streamer 8/2.
    """

    TIMEOUT = 10
    REPEAT_INFINITELY=-1
    AUTO=-1
    
    def __init__(self, ip_hostname='pulsestreamer'):
        print("Connect to Pulse Streamer via google-RPC.")
        print("IP / Hostname:", ip_hostname)
        self.ip_address = ip_hostname
        self.protocol = 'gRPC'
        self.client_release = __version__
        try:
            channel = insecure_channel(ip_hostname + ':50051')
            self.stub = PulseStreamerStub(channel)
            try:
                void = VoidMessage()
                self.stub.getSerial(void, timeout=PulseStreamer.TIMEOUT)
            except:
                try:
                    void = VoidMessage()
                    self.stub.isRunning(void, timeout=PulseStreamer.TIMEOUT)
                    assert False, "Pulse Streamer class not compatible with current firmware. Please update your firmware." \
                        "For detailed information visit https://www.swabianinstruments.com " \
                        "or contact support@swabianinstruments.com"
                except AssertionError:
                    raise
                except:
                    assert False, "No Pulse Streamer found at IP/Host-address: "+ip_hostname
        except AssertionError:
            raise
        except:
            assert False, "No Pulse Streamer found at IP/Host-address: "+ip_hostname
        void = VoidMessage()
        firmware_version=self.stub.getFirmwareVersion(void, timeout=PulseStreamer.TIMEOUT).string_value
        self.__version_1_1_or_higher=True if _compare_version_number(firmware_version, '1.1.0') >=0 else False
        print("Pulse Streamer 8/2 firmware: v" + firmware_version)
        print("Client software: v" + __version__)
        if(_compare_version_number(firmware_version) >= 0):
            print("The Pulse Streamer 8/2 firmware is more up to date than your client software. We recommend updating your client software.")
            print("For detailed information visit https://www.swabianinstruments.com/pulse-streamer-8-2/downloads/ or contact support@swabianinstruments.com")
        elif(_compare_version_number('.'.join(__version__.split('.')[0:2]+['0']),'.'.join(firmware_version.split('.')[0:2]+['0'])) > 0):
            print("Your client software is more up to date than the Pulse Streamer 8/2 firmware. We recommend updating the firmware of your Pulse Streamer 8/2.")
            print("For detailed information visit https://www.swabianinstruments.com/pulse-streamer-8-2/downloads/ or contact support@swabianinstruments.com")
        self._auto_slot=0
        self._has_data=False

    def reset(self):
        self._auto_slot=0
        self._has_data=False
        void = VoidMessage()
        return self.stub.reset(void, timeout=PulseStreamer.TIMEOUT).value

    def reboot(self):
        self._auto_slot=0
        self._has_data=False
        void = VoidMessage()
        return self.stub.reboot(void, timeout=PulseStreamer.TIMEOUT).value
        
    def constant(self, state=OutputState.ZERO()):
        if isinstance(state, OutputState):
            state=state.getData()
            sequence_step = PulseMessage(ticks=0, digi=state[0], ao0=state[1], ao1=state[2])
        else:
            if isinstance(state[0], list):
                sequence_step = self.convert_sequence_step((0, state[0], state[1], state[2]))
            else:
                chans, a0, a1 = state
                sequence_step = PulseMessage(ticks=0, digi=chans, ao0=a0, ao1=a1)
        self._auto_slot=0
        self._has_data=False
        self.stub.constant(sequence_step, timeout=PulseStreamer.TIMEOUT)

    def forceFinal(self):
        self._auto_slot=0
        self._has_data=False
        void = VoidMessage()
        return self.stub.forceFinal(void, timeout=PulseStreamer.TIMEOUT).value

    def createOutputState(self, digi, A0=0.0, A1=0.0):
        output=OutputState(digi=digi,A0=A0, A1=A1)
        return output

    def createSequence(self):
        seq = Sequence()
        return seq

    def stream(self, seq, n_runs=REPEAT_INFINITELY, final=OutputState.ZERO()):
        if (self.__version_1_1_or_higher):
            #print("Debug information: Stream binary")
            #return self.__stream_binary(seq, n_runs, final)
            self.__stream_binary(seq, n_runs, final)
        else:
            #print("Debug information: Stream grpc")
            self.__stream_grpc(seq, n_runs, final)
        
    def __stream_grpc(self, seq, n_runs=REPEAT_INFINITELY, final=OutputState.ZERO()):
        if isinstance(seq, Sequence):
            sequence=seq.getData()
        else:
            sequence= seq

        s = SequenceMessage()
        
        if type(sequence[0][1])== list:
            for i in range(len(sequence)):
                t, chans, a0, a1 = sequence[i]
                if t > 0xffffffff:
                    mod=t%0xffffffff
                    count=t//0xffffffff
                    for i in range(count):
                        t =0xffffffff
                        s.pulse.extend([self.convert_sequence_step((t, chans, a0, a1))])
                    else:
                        t=mod
                s.pulse.extend([self.convert_sequence_step((t, chans, a0, a1))])
        else:
            for i in range(len(sequence)):
                t, chans, a0, a1 = sequence[i]
                if t > 0xffffffff:
                    mod=t%0xffffffff
                    count=t//0xffffffff
                    for i in range(count):
                        t =0xffffffff
                        s.pulse.extend([PulseMessage(ticks=t, digi=chans, ao0=a0, ao1=a1)])
                    else:
                        t=mod
                s.pulse.extend([PulseMessage(ticks=t, digi=chans, ao0=a0, ao1=a1)])

        assert len(s.pulse)<=2e6, "The resulting sequence length exceeds the limit of two million sequnence steps"

        s.n_runs = n_runs

        if isinstance(final, OutputState):
            final=final.getData()
            conv_final = PulseMessage(ticks=0, digi=final[0],ao0=final[1], ao1=final[2])
        else:
            if isinstance(final[0], list):
                conv_final = self.convert_sequence_step((0,final[0], final[1], final[2]))
            else:
                chans, a0, a1 = final
                conv_final = PulseMessage(ticks=0, digi=chans, ao0=a0, ao1=a1)

        s.final.ticks = conv_final.ticks
        s.final.digi = conv_final.digi
        s.final.ao0 = conv_final.ao0
        s.final.ao1 = conv_final.ao1
  
        self.stub.stream(s, timeout=PulseStreamer.TIMEOUT)

    def __stream_binary(self, seq, n_runs=REPEAT_INFINITELY, final=OutputState.ZERO()):
        HOST = self.ip_address #'192.168.0.101' #''127.0.0.1'  # The server's hostname or IP address
        PORT = 21328        # The port used by the server
        MAGIC_TOKEN = 0x53504953
        COMMAND=0
        
        if isinstance(final, OutputState):
            state=final.getData()
            final =(0, state[0], state[1], state[2])
        else:
            final = self.convert_sequence_step((0, final[0], final[1], final[2]), binary=True)

        
        if isinstance(seq, Sequence):
            try:
                s_enc = self.enc_binary(seq.getData(as_ndarray=True))
                pulses=len(s_enc)//9
            except TypeError:
                s_enc = self.enc_binary(seq.getData())
                pulses=len(s_enc)//4
        else:
            s_enc = self.enc_binary(seq)
            pulses=len(s_enc)//4
        
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((HOST, PORT))
        pad=32-((pulses*9)%32)

        CommandHeader=[MAGIC_TOKEN, 1, COMMAND, 0x0, pulses*9+32+pad, 0]
        StreamHighLevelHeader=[n_runs, pulses, final[2], final[3], final[1], 0x0, 0x0, 0x0, 0]
        s.sendall(struct.pack('<4I2Q', *CommandHeader))
        s.sendall(struct.pack('<qQ2h4BQ', *StreamHighLevelHeader))
        if isinstance(s_enc, bytes):
            s.sendall(s_enc)
        else:
            fmt = '<' + len(s_enc)//4*'IBhh'
            s.sendall(struct.pack(fmt, *s_enc))
        s.sendall(b'\0' * pad)
        data = s.recv(32)
        s.close()


        return_value=struct.unpack('<4I2Q', data)
        return return_value

    def upload(self, slot_nr, data, n_runs=REPEAT_INFINITELY, idle_state=OutputState.ZERO(), next_action=NextAction.SWITCH_SLOT_EXPECT_NEW_DATA, when=When.IMMEDIATE, on_nodata=OnNoData.ERROR):
        HOST = self.ip_address #'192.168.0.101' #''127.0.0.1'  # The server's hostname or IP address
        PORT = 21328        # The port used by the server
        MAGIC_TOKEN = 0x53504953
        COMMAND=0x100

        if(slot_nr==self.AUTO): #handle auto-mode for slot_nr
            """Here we could check the real internal state via debug register to handle also cases, where the device was power-cycled,
            but this would cost performance (~2ms Linux/>10ms Windows) with auto-mode"""
            slot_nr=self._auto_slot 
        
        if isinstance(idle_state, OutputState):
            state=idle_state.getData()
            idle_state =(0, state[0], state[1], state[2])
        else:
            idle_state = self.convert_sequence_step((0, idle_state[0], idle_state[1], idle_state[2]), binary=True)

        if isinstance(data, Sequence):
            try:
                s_enc = self.enc_binary(data.getData(as_ndarray=True))
                pulses=len(s_enc)//9
            except TypeError:
                s_enc = self.enc_binary(data.getData())
                pulses=len(s_enc)//4
        else:
            s_enc = self.enc_binary(data)
            pulses=len(s_enc)//4
        
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((HOST, PORT))
        pad=32-((pulses*9)%32)

        CommandHeader=[MAGIC_TOKEN, 1, COMMAND, 0x0, pulses*9+32+pad, 0]
        UploadHeader=[n_runs, pulses, idle_state[2], idle_state[3], idle_state[1], 0x0, 0x0, 0x0, slot_nr, next_action.value, when.value, on_nodata.value, 0x0]
        s.sendall(struct.pack('<4I2Q', *CommandHeader))
        s.sendall(struct.pack('<qQ2h8BI', *UploadHeader))
        try:
            if isinstance(s_enc, bytes):
                s.sendall(s_enc)
            else:
                fmt = '<' + len(s_enc)//4*'IBhh'
                s.sendall(struct.pack(fmt, *s_enc))
            s.sendall(b'\0' * pad)
        except ConnectionResetError:
            data = s.recv(32)
            cmd_answer=struct.unpack('<4I2Q', data)
            s.close()
            if (cmd_answer[2]!=0):
                print("Binary transfer failed with error code: " + str(cmd_answer[2])) 
                raise
        data = s.recv(32)
        cmd_answer=struct.unpack('<4I2Q', data)
        assert cmd_answer[2]==0, "Binary transfer failed with error code: " + str(cmd_answer[2]) #End with error code if ConnectionResetError-Exception is not thrown
        data =s.recv(cmd_answer[4])
        return_value_upload=struct.unpack('<8i', data)
        s.close()

        if(return_value_upload[0]==0 and (next_action!=NextAction.SWITCH_SLOT or not self._has_data)): #handle auto-mode for slot_nr
            self._auto_slot=1-slot_nr
            self._has_data=True
        
        return return_value_upload[0]
            
    def start(self, slot_nr=0, slots_to_run=REPEAT_INFINITELY):
        if(slot_nr==self.AUTO): #auto-mode always starts with 0, otherwise the real state of the device has to be checked
            slot_nr=0
        s=StartMessage()
        s.slot_nr=slot_nr
        s.slots_to_run=slots_to_run
        return self.stub.start(s, timeout=PulseStreamer.TIMEOUT).value

    def isReadyForData(self, slot_nr=AUTO):
        if (slot_nr==self.AUTO): #handle auto-mode for slot_nr
            slot_nr=self._auto_slot
        d=ReadyForDataMessage()
        d.slot_nr=slot_nr
        return bool(self.stub.isReadyForData(d, timeout=PulseStreamer.TIMEOUT).value)
    
    def isStreaming(self):
        void = VoidMessage()
        return bool(self.stub.isStreaming(void, timeout=PulseStreamer.TIMEOUT).value)

    def hasSequence(self):
        void = VoidMessage()
        return bool(self.stub.hasSequence(void, timeout=PulseStreamer.TIMEOUT).value)

    def hasFinished(self):
        void = VoidMessage()
        return bool(self.stub.hasFinished(void, timeout=PulseStreamer.TIMEOUT).value)

    def startNow(self):
        void = VoidMessage()
        return self.stub.startNow(void, timeout=PulseStreamer.TIMEOUT).value

    def selectClock(self, source):
        if not isinstance(source, ClockSource):
            raise TypeError("source must be an instance of ClockSource Enum")
        else:
            c = ClockMessage()
            c.clock_source = source.value
            return self.stub.selectClock(c, timeout=PulseStreamer.TIMEOUT).value
    
    def getClock(self):
        void = VoidMessage()
        return ClockSource(self.stub.getClock(void, timeout=PulseStreamer.TIMEOUT).value)

    def getFirmwareVersion(self):
        void = VoidMessage()
        return self.stub.getFirmwareVersion(void, timeout=PulseStreamer.TIMEOUT).string_value

    def getHardwareVersion(self):
        void = VoidMessage()
        return self.stub.getHardwareVersion(void, timeout=PulseStreamer.TIMEOUT).string_value

    def flash_lic(self):
        void = VoidMessage()
        return self.stub.flash_lic(void, timeout=PulseStreamer.TIMEOUT).string_value

    def getSerial(self):
        if (self.__version_1_1_or_higher):
            void = VoidMessage()
            return self.stub.getSerial(void, timeout=PulseStreamer.TIMEOUT).string_value
        else:
            s=GetSerialMessage()
            s.serial = 1
            return self.stub.getSerial(s, timeout=PulseStreamer.TIMEOUT).string_value

    def getFPGAID(self):
        if (self.__version_1_1_or_higher):
            void = VoidMessage()
            return self.stub.getFPGAID(void, timeout=PulseStreamer.TIMEOUT).string_value
        else:
            s=GetSerialMessage()
            s.serial = 0
            return self.stub.getSerial(s, timeout=PulseStreamer.TIMEOUT).string_value

    def setClkParam(self, value=0):
        assert value in range(58)
        v= DrpClkSetMessage()
        v.value = value
        return self.stub.setClkParam(v, timeout=PulseStreamer.TIMEOUT).value

    def setTrigger(self, start, rearm=TriggerRearm.AUTO):
        if not isinstance(start, TriggerStart):
            raise TypeError("start must be an instance of TriggerStart Enum")
        else:
            if not isinstance(rearm, TriggerRearm):
                raise TypeError("rearm must be an instance of TriggerRearm Enum")
            else:
                t = TriggerMessage()
                t.start = start.value
                t.mode = rearm.value
                return self.stub.setTrigger(t, timeout=PulseStreamer.TIMEOUT).value

    def getTriggerStart(self):
        void = VoidMessage()
        return TriggerStart(self.stub.getTriggerStart(void, timeout=PulseStreamer.TIMEOUT).value)

    def getTriggerRearm(self):
        void = VoidMessage()
        return TriggerRearm(self.stub.getTriggerRearm(void, timeout=PulseStreamer.TIMEOUT).value)

    def rearm(self):
        void = VoidMessage()
        return bool(self.stub.rearm(void, timeout=PulseStreamer.TIMEOUT).value)

    def setAnalogCalibration(self, dc_offset_a0=0, dc_offset_a1=0, slope_a0=1, slope_a1=1):
        a=AnalogCalibrationData()
        a.dc_offset_a0=dc_offset_a0
        a.dc_offset_a1=dc_offset_a1
        a.slope_a0=slope_a0
        a.slope_a1=slope_a1
        return self.stub.setAnalogCalibration(a,timeout=PulseStreamer.TIMEOUT).value
    
    def getAnalogCalibration(self):
        void = VoidMessage()
        response = self.stub.getAnalogCalibration(void, timeout=PulseStreamer.TIMEOUT)
        key_list=['dc_offset_a0', 'dc_offset_a1', 'slope_a0', 'slope_a1']
        return dict( zip(key_list, [response.dc_offset_a0, response.dc_offset_a1, response.slope_a0, response.slope_a1] ))

    def getSupplyState(self):
        void = VoidMessage()
        response = self.stub.getSupplyState(void, timeout=PulseStreamer.TIMEOUT)
        key_list=['Current/A', 'Voltage_12/V', 'Voltage_3.3/V']
        return dict( zip(key_list, [response.efuse_current, response.v12, response.v33] ))

    def setSquareWave125MHz(self, channels=[]):
        s=SetSquareWaveMessage()
        s.digi = self.chans_to_mask(channels)
        return self.stub.setSquareWave125MHz(s, timeout=PulseStreamer.TIMEOUT).value
    
    def setNetworkConfiguration(self, dhcp, ip='', netmask='', gateway='', testmode=True):
        assert dhcp in [True, False]
        assert testmode in [True, False]
        n=SetNetworkMessage()
        n.dhcp=dhcp
        n.ip=ip
        n.netmask=netmask
        n.gateway=gateway
        n.testmode=testmode
        return self.stub.setNetworkConfiguration(n, timeout=PulseStreamer.TIMEOUT).value

    def getNetworkConfiguration(self, permanent=False):
        assert permanent in [True, False]
        g=GetNetworkMessage()
        g.permanent=permanent
        response = self.stub.getNetworkConfiguration(g, timeout=PulseStreamer.TIMEOUT)
        key_list=['dhcp', 'ip', 'netmask', 'gateway']
        return dict( zip(key_list, [response.dhcp, response.ip, response.netmask, response.gateway]))

    def applyNetworkConfiguration(self):
        void = VoidMessage()
        return self.stub.applyNetworkConfiguration(void, timeout=PulseStreamer.TIMEOUT).value

    def setHostname(self, hostname):
        h=HostnameMessage()
        h.hostname=hostname
        return self.stub.setHostname(h, timeout=PulseStreamer.TIMEOUT).value

    def getHostname(self):
        void = VoidMessage()
        return self.stub.getHostname(void, timeout=PulseStreamer.TIMEOUT).string_value

    def getTemperature(self):
        void = VoidMessage()
        return self.stub.getTemperature(void, timeout=PulseStreamer.TIMEOUT).string_value

    def getUnderflow(self):
        void = VoidMessage()
        return self.stub.getUnderflow(void, timeout=PulseStreamer.TIMEOUT).value

    def getDebugRegister(self):
        void = VoidMessage()
        return self.stub.getDebugRegister(void, timeout=PulseStreamer.PulseStreamer.TIMEOUT).value

    def chans_to_mask(self, chans):
        mask = 0
        for chan in chans:
            mask |= 1<<chan
        assert mask <= 255, "Pulse Streamer 8/2 supports "\
            "up to eight digital channels"
        return mask

    def convert_sequence_step(self, sequence_step, binary=False):
        t, chans, a0, a1 = sequence_step
        assert (abs(a0)<=1 and abs(a1)<=1), "Pulse Streamer 8/2 supports "\
                "analog voltage range of +/-1V" #check hardware
        assert t>=0
        if binary:
            return (t, self.chans_to_mask(chans), int(round(0x7fff*a0)), int(round(0x7fff*a1)))
        else:
            return PulseMessage(ticks=t, digi=self.chans_to_mask(chans), ao0=int(round(0x7fff*a0)), ao1=int(round(0x7fff*a1)))

    def enc_binary(self, seq):
        """
        Convert a human readable python sequence to a base64 encoded string and split sequence steps with duration exceeding limit of unsigned int
        """
        s = b''
        convert_list = []
        if isinstance(seq, np.ndarray):
            elements_per_pulse=9
            convert_list=seq.tobytes('A')
        elif type(seq[0][1])== list:
            for sequence_step in seq:
                if sequence_step[0] > 0xffffffff:
                    mod=sequence_step[0]%0xffffffff
                    count=sequence_step[0]//0xffffffff
                    for i in range(count):
                        sequence_step = (0xffffffff, sequence_step[1], sequence_step[2], sequence_step[3])
                        convert_list.extend(self.convert_sequence_step(sequence_step, binary=True))
                    else:
                        sequence_step = (mod, sequence_step[1], sequence_step[2], sequence_step[3])
                convert_list.extend(self.convert_sequence_step(sequence_step, binary=True))
            elements_per_pulse=4    
        else:
            for sequence_step in seq:
                if sequence_step[0] > 0xffffffff:
                    mod=sequence_step[0]%0xffffffff
                    count=sequence_step[0]//0xffffffff
                    for i in range(count):
                        sequence_step = (0xffffffff, sequence_step[1], sequence_step[2], sequence_step[3])
                        convert_list.extend(sequence_step)
                    else:
                        sequence_step = (mod, sequence_step[1], sequence_step[2], sequence_step[3])
                convert_list.extend(sequence_step)
            elements_per_pulse=4

        assert len(convert_list)//elements_per_pulse<=2e6, "The resulting sequence length exceeds the limit of two million sequnence steps"

        return convert_list

"""---------Test-Code-------------------------------"""

if __name__ == '__main__':
    pulser = PulseStreamer(ip_hostname='pulsestreamer')

    print("Serial number:", pulser.getSerial())
    print("Firmware Version:", pulser.getFirmwareVersion())

    